@@ -26,15 +26,47 @@
  * If you do not delete the provisions above, a recipient may use your
  * version of this file under either the License or the GPL.
  *
- * $Header: /cvsroot/aolserver/nspostgres/nspostgres.c,v 1.7 2003/06/28 19:15:47 seryakov Exp $
+ * $Header: /cvsroot/aolserver/nspostgres/nspostgres.c,v 1.15 2007/02/25 09:45:32 jwl Exp $
  */
 
 #include "nspostgres.h"
+#include <tcl.h>
 
 DllExport int   Ns_ModuleVersion = 1;
 
 static char datestyle[STRING_BUF_LEN];
 
+
+/* 
+ * append_PQresultStatus
+ *
+ * accepts a pointer to a fully prepared and usable Tcl_DString
+ * and appends to it a descriptive string stating the status of the 
+ * incoming PQresult.
+ *
+ * returns nothing; the collector string is an in-out parameter.
+ * As such, it could already contain a string at the point of this
+ * function call.
+ *
+ */
+
+void
+append_PQresultStatus(Tcl_DString *collector, const PGresult *pgResult)
+{
+  Tcl_DStringAppend(collector, "(Status of PQexec call: ", -1);
+  
+  if(pgResult)
+    {
+      Tcl_DStringAppend(collector, PQresStatus(PQresultStatus(pgResult)), -1);
+    }
+  else
+    {
+      Tcl_DStringAppend(collector, "none; PQexec returned null pointer", -1);
+    }
+  
+  Tcl_DStringAppend(collector, ")\n", -1);
+}
+
 DllExport int
 Ns_DbDriverInit(char *hDriver, char *configPath)
 {
@@ -458,7 +490,7 @@ Ns_PgSelect(Ns_DbHandle *handle, char *sql) {
 
     static char *asfuncname = "Ns_PgSelect";
     Ns_Set         *row = NULL;
-    NsPgConn       *nsConn;
+    NsPgConn       *nspgConn;
     int             i;
 
     if (handle == NULL || handle->connection == NULL) {
@@ -471,24 +503,59 @@ Ns_PgSelect(Ns_DbHandle *handle, char *sql) {
         goto done;
     }
 
-    nsConn = handle->connection;
+    nspgConn = handle->connection;
 
     if (Ns_PgExec(handle, sql) != NS_ERROR) {
 
-        if (PQresultStatus(nsConn->res) == PGRES_TUPLES_OK) {
-            nsConn->curTuple = 0;
-            nsConn->nCols = PQnfields(nsConn->res);
-            nsConn->nTuples = PQntuples(nsConn->res);
+        if (PQresultStatus(nspgConn->res) == PGRES_TUPLES_OK) {
+            nspgConn->curTuple = 0;
+            nspgConn->nCols = PQnfields(nspgConn->res);
+            nspgConn->nTuples = PQntuples(nspgConn->res);
             row = handle->row;
             
-            for (i = 0; i < nsConn->nCols; i++) {
-                Ns_SetPut(row, (char *)PQfname(nsConn->res, i), NULL);
+            for (i = 0; i < nspgConn->nCols; i++) {
+                Ns_SetPut(row, (char *)PQfname(nspgConn->res, i), NULL);
             }
             
         } else {
-            Ns_Log(Error, "Ns_PgSelect(%s):  Query did not return rows:  %s",
-                   handle->datasource, sql);
+	  Tcl_DString errString;
+	  Tcl_DStringInit(&errString);
+
+	  Tcl_DStringAppend(&errString, "\nNs_PgSelect(", -1);
+	  Tcl_DStringAppend(&errString, handle->datasource, -1);
+	  Tcl_DStringAppend(&errString, "):  Query did not return rows\n", -1);
+
+	  if(handle->verbose)
+	    {
+	      append_PQresultStatus(&errString, nspgConn->res);
+
+	      Tcl_DStringAppend(&errString, "SQL:  ", -1);
+	      Tcl_DStringAppend(&errString, sql, -1);
+	    }
+
+	  Ns_Log(Error, "%s", Tcl_DStringValue(&errString));
+
+	  Tcl_DStringFree(&errString);
         }
+    } else {
+      // Ns_PgExec returned NS_ERROR
+
+      Tcl_DString errString;
+      Tcl_DStringInit(&errString);
+      
+      Tcl_DStringAppend(&errString, "\nNs_PgSelect(", -1);
+      Tcl_DStringAppend(&errString, handle->datasource, -1);
+      Tcl_DStringAppend(&errString, "):  Ns_PgExec returned NS_ERROR\n", -1);
+      
+      if(handle->verbose)
+	{
+	  append_PQresultStatus(&errString, nspgConn->res);
+	  
+	  Tcl_DStringAppend(&errString, "SQL:  ", -1);
+	  Tcl_DStringAppend(&errString, sql, -1);
+	}
+      
+      Ns_Log(Error, "%s", Tcl_DStringValue(&errString));
     }
   done:
     return (row);
@@ -503,7 +570,7 @@ static int
 Ns_PgGetRow(Ns_DbHandle *handle, Ns_Set *row) {
 
     static char    *asfuncname = "Ns_PgGetRow";
-    NsPgConn       *nsConn;
+    NsPgConn       *nspgConn;
     int             i;
 
     if (handle == NULL || handle->connection == NULL) {
@@ -516,25 +583,25 @@ Ns_PgGetRow(Ns_DbHandle *handle, Ns_Set *row) {
         return NS_ERROR;
     }
 
-    nsConn = handle->connection;
+    nspgConn = handle->connection;
 
-    if (nsConn->nCols == 0) {
+    if (nspgConn->nCols == 0) {
         Ns_Log(Error, "Ns_PgGetRow(%s):  Get row called outside a fetch row loop.",
                handle->datasource);
         return NS_ERROR;
-    } else if (nsConn->curTuple == nsConn->nTuples) {
+    } else if (nspgConn->curTuple == nspgConn->nTuples) {
 
-        PQclear(nsConn->res);
-        nsConn->res = NULL;
-        nsConn->nCols = nsConn->nTuples = nsConn->curTuple = 0;
+        PQclear(nspgConn->res);
+        nspgConn->res = NULL;
+        nspgConn->nCols = nspgConn->nTuples = nspgConn->curTuple = 0;
         return NS_END_DATA;
 
     } else {
-        for (i = 0; i < nsConn->nCols; i++) {
-            Ns_SetPutValue(row, i, (char *) PQgetvalue(nsConn->res,
-				 nsConn->curTuple, i));
+        for (i = 0; i < nspgConn->nCols; i++) {
+            Ns_SetPutValue(row, i, (char *) PQgetvalue(nspgConn->res,
+				 nspgConn->curTuple, i));
         }
-        nsConn->curTuple++;
+        nspgConn->curTuple++;
     }
 
     return NS_OK;
@@ -548,19 +615,19 @@ static int
 Ns_PgFlush(Ns_DbHandle *handle) {
 
     static char *asfuncname = "Ns_PgFlush";
-    NsPgConn   *nsConn;
+    NsPgConn   *nspgConn;
 
     if (handle == NULL || handle->connection == NULL) {
         Ns_Log(Error, "%s: Invalid connection.", asfuncname);
         return NS_ERROR;
     } 
 
-    nsConn = handle->connection;
+    nspgConn = handle->connection;
 
-    if (nsConn->nCols > 0) {
-        PQclear(nsConn->res);
-        nsConn->res = NULL;
-        nsConn->nCols = nsConn->nTuples = nsConn->curTuple = 0;
+    if (nspgConn->nCols > 0) {
+        PQclear(nspgConn->res);
+        nspgConn->res = NULL;
+        nspgConn->nCols = nspgConn->nTuples = nspgConn->curTuple = 0;
     }
     return NS_OK;
 }
@@ -789,7 +856,7 @@ decode3(unsigned char *p, char *buf, int n)
 static int
 blob_get(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id)
 {
-    NsPgConn	*nsConn = (NsPgConn *) handle->connection;
+    NsPgConn	*nspgConn = (NsPgConn *) handle->connection;
 	int			segment;
 	char		query[100];
 	char		*segment_pos;
@@ -810,15 +877,33 @@ blob_get(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id)
 		char	buf[6001];
 
 		sprintf(segment_pos, "%d", segment);
+
 		if (Ns_PgExec(handle, query) != NS_ROWS) {
-			Tcl_AppendResult(interp, "Error selecting data from BLOB", NULL);
-			return TCL_ERROR;
+		  Tcl_DString errString;
+		  Tcl_DStringInit(&errString);
+		  
+		  Tcl_DStringAppend
+		    (&errString, "Error selecting data from BLOB\n", -1);
+		  
+		  if(handle->verbose)
+		    {
+		      append_PQresultStatus(&errString, nspgConn->res);
+		      
+		      Tcl_DStringAppend(&errString, "SQL:  ", -1);
+		      Tcl_DStringAppend(&errString, query, -1);
+		    }
+
+		  Tcl_AppendResult(interp, Tcl_DStringValue(&errString), NULL);
+		  
+		  Tcl_DStringFree(&errString);
+		  
+		  return TCL_ERROR;
 		}
 
-		if (PQntuples(nsConn->res) == 0) break;
+		if (PQntuples(nspgConn->res) == 0) break;
 
-		byte_len_column = PQgetvalue(nsConn->res, 0, 0);
-		data_column = PQgetvalue(nsConn->res, 0, 1);
+		byte_len_column = PQgetvalue(nspgConn->res, 0, 0);
+		data_column = PQgetvalue(nspgConn->res, 0, 1);
 		sscanf(byte_len_column, "%d", &byte_len);
 		nbytes += byte_len;
 		n = byte_len;
@@ -830,8 +915,8 @@ blob_get(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id)
 		segment++;
     }
 
-	PQclear(nsConn->res);
-	nsConn->res = NULL;
+	PQclear(nspgConn->res);
+	nspgConn->res = NULL;
 
 	return TCL_OK;
 }
@@ -877,6 +962,7 @@ static int
 blob_put(Tcl_Interp *interp, Ns_DbHandle *handle, char* blob_id,
 			char* value)
 {
+    NsPgConn	*nspgConn = (NsPgConn *) handle->connection;
 	int			i, j, segment, value_len, segment_len;
 	char		out_buf[8001], query[10000];
 	char		*segment_pos, *value_ptr;
@@ -898,9 +984,27 @@ blob_put(Tcl_Interp *interp, Ns_DbHandle *handle, char* blob_id,
 		}
 		out_buf[j] = '\0';
 		sprintf(segment_pos, "%d, %d, '%s')", segment, segment_len, out_buf);
+
 		if (Ns_PgExec(handle, query) != NS_DML) {
-			Tcl_AppendResult(interp, "Error inserting data into BLOB", NULL);
-			return TCL_ERROR;
+		  Tcl_DString errString;
+		  Tcl_DStringInit(&errString);
+		  
+		  Tcl_DStringAppend
+		    (&errString, "Error inserting data into BLOB\n", -1);
+		  
+		  if(handle->verbose)
+		    {
+		      append_PQresultStatus(&errString, nspgConn->res);
+		      
+		      Tcl_DStringAppend(&errString, "SQL:  ", -1);
+		      Tcl_DStringAppend(&errString, query, -1);
+		    }
+
+		  Tcl_AppendResult(interp, Tcl_DStringValue(&errString), NULL);
+		  
+		  Tcl_DStringFree(&errString);
+		  
+		  return TCL_ERROR;
 		}
         value_ptr += segment_len;
 		segment++;
@@ -917,6 +1021,7 @@ static int
 blob_dml_file(Tcl_Interp *interp, Ns_DbHandle *handle, char* blob_id,
 			char* filename)
 {
+    NsPgConn	*nspgConn = (NsPgConn *) handle->connection;
 	int			fd, i, j, segment, readlen;
 	char		in_buf[6000], out_buf[8001], query[10000];
 	char		*segment_pos;
@@ -946,8 +1051,25 @@ blob_dml_file(Tcl_Interp *interp, Ns_DbHandle *handle, char* blob_id,
 		out_buf[j] = '\0';
 		sprintf(segment_pos, "%d, %d, '%s')", segment, readlen, out_buf);
 		if (Ns_PgExec(handle, query) != NS_DML) {
-			Tcl_AppendResult(interp, "Error inserting data into BLOB", NULL);
-			return TCL_ERROR;
+		  Tcl_DString errString;
+		  Tcl_DStringInit(&errString);
+		  
+		  Tcl_DStringAppend
+		    (&errString, "Error inserting data into BLOB\n", -1);
+		  
+		  if(handle->verbose)
+		    {
+		      append_PQresultStatus(&errString, nspgConn->res);
+		      
+		      Tcl_DStringAppend(&errString, "SQL:  ", -1);
+		      Tcl_DStringAppend(&errString, query, -1);
+		    }
+
+		  Tcl_AppendResult(interp, Tcl_DStringValue(&errString), NULL);
+		  
+		  Tcl_DStringFree(&errString);
+		  
+		  return TCL_ERROR;
 		}
 		readlen = read(fd, in_buf, 6000);
 		segment++;
@@ -974,7 +1096,7 @@ static int
 blob_send_to_stream(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id, 
 		    int to_conn_p, char* filename)
 {
-  NsPgConn	*nsConn = (NsPgConn *) handle->connection;
+  NsPgConn	*nspgConn = (NsPgConn *) handle->connection;
   int		segment;
   char		query[100];
   int		fd;
@@ -1031,14 +1153,31 @@ blob_send_to_stream(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id,
 
     sprintf(segment_pos, "%d", segment);
     if (Ns_PgExec(handle, query) != NS_ROWS) {
-      Tcl_AppendResult(interp, "Error selecting data from BLOB", NULL);
+      Tcl_DString errString;
+      Tcl_DStringInit(&errString);
+      
+      Tcl_DStringAppend
+	(&errString, "Error selecting data from BLOB\n", -1);
+      
+      if(handle->verbose)
+	{
+	  append_PQresultStatus(&errString, nspgConn->res);
+	  
+	  Tcl_DStringAppend(&errString, "SQL:  ", -1);
+	  Tcl_DStringAppend(&errString, query, -1);
+	}
+      
+      Tcl_AppendResult(interp, Tcl_DStringValue(&errString), NULL);
+      
+      Tcl_DStringFree(&errString);
+      
       return TCL_ERROR;
     }
 
-    if (PQntuples(nsConn->res) == 0) break;
+    if (PQntuples(nspgConn->res) == 0) break;
 
-    byte_len_column = PQgetvalue(nsConn->res, 0, 0);
-    data_column = PQgetvalue(nsConn->res, 0, 1);
+    byte_len_column = PQgetvalue(nspgConn->res, 0, 0);
+    data_column = PQgetvalue(nspgConn->res, 0, 1);
     sscanf(byte_len_column, "%d", &byte_len);
     n = byte_len;
     for (i=0, j=0; n > 0; i += 4, j += 3, n -= 3) {
@@ -1055,8 +1194,8 @@ blob_send_to_stream(Tcl_Interp *interp, Ns_DbHandle *handle, char* lob_id,
       close (fd);
     }
 
-  PQclear(nsConn->res);
-  nsConn->res = NULL;
+  PQclear(nspgConn->res);
+  nspgConn->res = NULL;
 
   return TCL_OK;
 }
@@ -1113,27 +1252,41 @@ BadArgs(Tcl_Interp *interp, char **argv, char *args)
 static int
 DbFail(Tcl_Interp *interp, Ns_DbHandle *handle, char *cmd, char* sql)
 {
-  NsPgConn *pgconn = handle->connection;
+  NsPgConn *nspgConn = handle->connection;
   char     *pqerror;
 
-  Tcl_AppendResult(interp, "Database operation \"", cmd, "\" failed", NULL);
+  Tcl_AppendResult(interp, "Database operation \"", cmd, "\" failed\n", NULL);
   if (handle->cExceptionCode[0] != '\0') {
-    Tcl_AppendResult(interp, " (exception ", handle->cExceptionCode,
+    Tcl_AppendResult(interp, "(exception ", handle->cExceptionCode,
                      NULL);
     if (handle->dsExceptionMsg.length > 0) {
       Tcl_AppendResult(interp, ", \"", handle->dsExceptionMsg.string,
                        "\"", NULL);
     }
-    Tcl_AppendResult(interp, ")", NULL);
+    Tcl_AppendResult(interp, ")\n", NULL);
   }
 
-  pqerror = PQerrorMessage(pgconn->conn);
-  if (pqerror[0] != '\0') {
-    Tcl_AppendResult(interp, "\n\n", pqerror, NULL);
-  } else {
-    Tcl_AppendResult(interp, "\n", NULL);
-  }
-  Tcl_AppendResult(interp, "\nSQL: ", sql, NULL);
+  if(handle->verbose)
+    {
+      pqerror = PQerrorMessage(nspgConn->conn);
+      if (pqerror != NULL && pqerror[0] != '\0') {
+	Tcl_AppendResult(interp, "pqerror was: \"", pqerror, "\"\n", NULL);
+      }
+
+      /*
+       * append the result status
+       */
+  
+      Tcl_DString pqString;
+      Tcl_DStringInit(&pqString);
+      
+      append_PQresultStatus(&pqString, nspgConn->res); 
+      
+      Tcl_AppendResult(interp, Tcl_DStringValue(&pqString), NULL);
+      Tcl_DStringFree(&pqString);
+
+      Tcl_AppendResult(interp, "SQL: ", sql, "\n", NULL);
+    }
 
   Ns_Free(sql);
 
@@ -1521,13 +1674,13 @@ static int
 PgCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv) {
 
     Ns_DbHandle    *handle;
-    NsPgConn        *pgconn;
+    NsPgConn        *nspgConn;
 
     if (Ns_TclDbGetHandle(interp, argv[2], &handle) != TCL_OK) {
         return TCL_ERROR;
     }
 
- 	pgconn = (NsPgConn *) handle->connection;
+ 	nspgConn = (NsPgConn *) handle->connection;
 
     /*
      * Make sure this is a PostgreSQL handle before accessing
@@ -1562,7 +1715,7 @@ PgCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv) {
                          	argv[0], " command dbId blobId value\"", NULL);
         	return TCL_ERROR;
     	}
-		if (!pgconn->in_transaction) {
+		if (!nspgConn->in_transaction) {
         	Tcl_AppendResult(interp,
 							 "blob_put only allowed in transaction", NULL);
         	return TCL_ERROR;
@@ -1574,7 +1727,7 @@ PgCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv) {
                          	argv[0], " command dbId blobId filename\"", NULL);
         	return TCL_ERROR;
     	}
-		if (!pgconn->in_transaction) {
+		if (!nspgConn->in_transaction) {
         	Tcl_AppendResult(interp,
 							 "blob_dml_file only allowed in transaction", NULL);
         	return TCL_ERROR;
@@ -1598,27 +1751,27 @@ PgCmd(ClientData dummy, Tcl_Interp *interp, int argc, char **argv) {
     }
 
     if (!strcmp(argv[1], "db")) {
-        Tcl_SetResult(interp, (char *) PQdb(pgconn->conn), TCL_STATIC);
+        Tcl_SetResult(interp, (char *) PQdb(nspgConn->conn), TCL_STATIC);
     } else if (!strcmp(argv[1], "host")) {
-        Tcl_SetResult(interp, (char *) PQhost(pgconn->conn), TCL_STATIC);
+        Tcl_SetResult(interp, (char *) PQhost(nspgConn->conn), TCL_STATIC);
     } else if (!strcmp(argv[1], "options")) {
-        Tcl_SetResult(interp, (char *) PQoptions(pgconn->conn), TCL_STATIC);
+        Tcl_SetResult(interp, (char *) PQoptions(nspgConn->conn), TCL_STATIC);
     } else if (!strcmp(argv[1], "port")) {
-        Tcl_SetResult(interp, (char *) PQport(pgconn->conn), TCL_STATIC);
+        Tcl_SetResult(interp, (char *) PQport(nspgConn->conn), TCL_STATIC);
     } else if (!strcmp(argv[1], "number")) {
-        sprintf(interp->result, "%u", pgconn->cNum);
+        sprintf(interp->result, "%u", nspgConn->cNum);
     } else if (!strcmp(argv[1], "error")) {
-        Tcl_SetResult(interp, (char *) PQerrorMessage(pgconn->conn),
+        Tcl_SetResult(interp, (char *) PQerrorMessage(nspgConn->conn),
 			 TCL_STATIC);
     } else if (!strcmp(argv[1], "status")) {
-        if (PQstatus(pgconn->conn) == CONNECTION_OK) {
+        if (PQstatus(nspgConn->conn) == CONNECTION_OK) {
             interp->result = "ok";
         } else {
             interp->result = "bad";
         }
     } else if (!strcmp(argv[1], "ntuples")) {
 	char string[16];
-	sprintf(string, "%d", pgconn->nTuples);
+	sprintf(string, "%d", nspgConn->nTuples);
 	Tcl_SetResult(interp, string, TCL_VOLATILE);
     } else {
         Tcl_AppendResult(interp, "unknown command \"", argv[2],
@@ -1790,6 +1943,8 @@ pg_column_command (ClientData dummy, Tcl_Interp *interp,
 
     if (argc < 4) 
       {
+	/* NOTE, the tcl gurus suggested Tcl_WrongNumArgs() for this */
+
 	Tcl_AppendResult (interp, "wrong # args:  should be \"",
 			  argv[0], " command dbId table ?args?\"", NULL);
 	goto bailout;
